// $License: NOLICENSE
//--------------------------------------------------------------------------------
/**
  @file $relPath
  @copyright $copyright
  @author mmarchhart
*/

//--------------------------------------------------------------------------------
// Libraries used (#uses)
#uses "classes/bsa/BsaControl"
#uses "classes/bsa/BsaWorkstation"
#uses "classes/bsa/AKS"
#uses "classes/bsa/BleSupervision"

//--------------------------------------------------------------------------------
// Variables and Constants

//--------------------------------------------------------------------------------
/**
*/

//the constance are only for parts where the ctrl languages does not allow the usage of the below defined enums (e.g. in switch statements)
const int MODE_SERVE_LOCAL = 11;
const int MODE_SERVE_REVISION = 21;
const int MODE_SERVE_INVALID = 254;

const int MODE_CONTROL_MANUAL = 1;
const int MODE_CONTROL_PROGRAM = 2;
const int MODE_CONTROL_MANUAL_TEMP = 3;

const string INFOTEXT_CONFIG = ":_general.._string_01";

enum BsaModeServe
{
  BLE = 0,
  BSA = 1,
  LOCAL = 11,
  REVISION = 21,
  INVALID = 254
};

enum BsaModeControl
{
  AUTOMATIC = 0,
  MANUAL = 1,
  PROGRAM = 2,
  MANUAL_TEMP = 3
};

enum BsaSpez
{
  NORMAL = 0,
  BLOCKED = 1,
  INACTIVE = 2,
  WINTER = 3,
  PRIORITY = 4
};

class BsaObject
{
//--------------------------------------------------------------------------------
//@public members
//--------------------------------------------------------------------------------

  //------------------------------------------------------------------------------
  /** The Default Constructor.
  */
  public BsaObject(const string &objectDpName)
  {
    _objectDpName = objectDpName;

    _workstation = BsaWorkstation::GetInstance();

    _bsaControl = _workstation.GetCurrentBsaControl();

    dpConnect(this, ModeServeControlCB, _objectDpName + ".cmdio.mode_serve",
                                        _objectDpName + ".cmdio.mode_control",
                                        _objectDpName + ".cmdio.spez");

    uiConnect(this, BSAControlChangedCB, _workstation.GetWorkstationShape(), "BSAControlChanged");
    uiConnect(this, BLESupervisionChangedCB, _workstation.GetWorkstationShape(), "BLESupervisionChanged");
  }

  /**
    Retruns the stored info text for the bsa object
    @return
  */
  public string GetInfoText()
  {
    string text;
    dpGet(_objectDpName + "." + INFOTEXT_CONFIG, text);

    return text;
  }

  /**
    Saves the given info text for the bsa object
    @param infoText
  */
  public void SetInfoText(const string &infoText)
  {
    dpSet(_objectDpName + "." + INFOTEXT_CONFIG, infoText);
  }

  /**
    Returns the Datapoint name of the object
    @return
  */
  public string GetObjectDpName()
  {
    return _objectDpName;
  }

   /**
    Returns the bsa object AKS
    @return
  */
  public shared_ptr<AKS> GetObjectAKS()
  {
    string aksString = dpGetAlias(_objectDpName + ".");
    shared_ptr<AKS> aks = AKS::Parse(aksString);  //mma: try to parse the string to an AKS object

    return aks;
  }

  /**
    Returns the bsa object AKS formated for display purposes (e.g. in title bars)
    @return
  */
  public string GetObjectAKSForTitle()
  {
    shared_ptr<AKS> aks = GetObjectAKS();

    string aksForTitle;

    if(aks != nullptr)
    {
      aksForTitle += aks.GetErfassungsbereich() + " ";
      aksForTitle += aks.GetStreckenabschnitt() + " ";
      aksForTitle += aks.GetAbschnitt() + " = ";
      aksForTitle += aks.GetAnlage() + " ";
      aksForTitle += aks.GetAnlagenteil() + " ";
      aksForTitle += aks.GetGruppe() + " ";
      aksForTitle += aks.GetObjekt();

      aksForTitle.replace("x", "");  //mma: remove the placehoder 'x'es from the string to omit them in the titles
    }
    else
    {
      aksForTitle = "AKSParsingError";
    }

    return aksForTitle;
  }

  public void SetModeControl(BsaModeControl modeControl)
  {
    dpSet(_objectDpName + ".cmdio.mode_control", (int)modeControl);
  }

  /**
    Registers a listener to changes on the modeControl of the bsa object
     The reference panel must implement public void ModeControlChangedCB(BsaModeControl modeControl, bool controlAllowed) to be a valid listener
    @param listener
    @return
  */
  public void RegisterModeControlChangeListener(shape listener)
  {
    if(listener != nullptr)
    {
      synchronized(_controlChangeListenerMutex)
      {
          _modeControlChangeListener.append(listener);
          listener.ModeControlChangedCB(_currentModeControl, _modeControlAllowed);  //send the initial values to the listener
      }
    }
  }

  /**
    Registers a listener to changes on the modeServe of the bsa object
     The reference panel must implement public void ModeServeChangedCB(BsaModeServe modeServe) to be a valid listener
    @param listener
    @return
  */
  public bool RegisterModeServeChangeListener(shape listener)
  {
    if(listener != nullptr)
    {
      synchronized(_serveChangeListenerMutex)
      {
        _modeServeChangeListener.append(listener);
        listener.ModeServeChangedCB(_currentModeServe);  //send the initial value to the listener
      }
    }
  }

  /**
    Registers a listener to changes on the spez of the bsa object
     The reference panel must implement public void SpezChangedCB(BsaSpez spez) to be a valid listener
    @param listener
    @return
  */
  public bool RegisterSpezChangeListener(shape listener)
  {
    if(listener != nullptr)
    {
      synchronized(_spezChangeListenerMutex)
      {
        _spezChangeListener.append(listener);
        listener.SpezChangedCB(_currentSpez);  //send the initial value to the listener
      }
    }
  }

  /**
    Returns true if the overall mode serve is in remote. one of the two (BLE or BSA)
    @return
  */
  public bool IsModeServeRemote()
  {
    switch(_currentModeServe)
    {
      case BsaModeServe::BLE :
      case BsaModeServe::BSA :
        return true;
        break;
      default :
        return false;
    }
  }

//--------------------------------------------------------------------------------
//@public static members
//--------------------------------------------------------------------------------
  /**
    connects to the given datapoints and show the fore/backcolor on the given shape
    @param dsDps ... datapoint instances
    @param dsDpes ... datapoint elements
    @param bShowAlertAsForeColor
    @param bShowAlertAsBackColor
    @param bShapeInvisibleIfNoAlert
    @param sShape ... shape
    @param iMinPrio ... minimalPriority to show
  */
  public static void showAlertColorOnShape(dyn_string dsDps, dyn_string dsDpes, string sShape,
                                           bool bShowAlertAsForeColor = true, bool bShowAlertAsBackColor = false, bool bShapeInvisibleIfNoAlert = true, int iMinPrio = 0)
  {
    dyn_string dsFrom;

    dynAppend(dsFrom, dsDps);
    dynAppend(dsFrom, dsDpes);

    for ( int i = 1; i <= dynlen(dsFrom); i++ )
    {

      if ( !dpExists(dsFrom[i]) )
      {
        setMultiValue(sShape, "foreCol", "_dpdoesnotexist",
                      sShape, "backCol", "_dpdoesnotexist",
                      sShape, "visible", true);
        return;
      }
    }

    string sFrom = dynStringToString(dsFrom, ",");

    mapping mUserData;

    mUserData["from"] = sFrom;
    mUserData["shape"] = sShape;
    mUserData["minPrio"] = iMinPrio;
    mUserData["bShowAlertAsForeColor"] = bShowAlertAsForeColor;
    mUserData["bShowAlertAsBackColor"] = bShowAlertAsBackColor;
    mUserData["bShapeInvisibleIfNoAlert"] = bShapeInvisibleIfNoAlert;

    dpQueryConnectSingle("showAlertColorOnShapeCB", true, mUserData,
                         "SELECT ALERT '_alert_hdl.._alert_color', '_alert_hdl.._value', '_alert_hdl.._prior', '_alert_hdl.._last_of_all', '_alert_hdl.._act_state', '_alert_hdl.._sum' "+
                         "FROM '{"+sFrom+"}' WHERE '_alert_hdl.._last_of_all' == 1 AND '_alert_hdl.._sum' == 0 AND '_alert_hdl.._prior' >= "+iMinPrio+" SORT BY 3 DESC");

  }

  /**
    Returns the object aks code from the given object datapoint name
    @param objectDpName
    @return
  */
  public static shared_ptr<AKS> GetAKSFromObjectDP(const string &objectDpName)
  {
    string aksString = dpGetAlias(objectDpName + ".");
    shared_ptr<AKS> aks = AKS::Parse(aksString);  //mma: try to parse the string to an AKS object

    return aks;
  }
//--------------------------------------------------------------------------------
//@protected members
//--------------------------------------------------------------------------------
  /**
    This method can be overwritten by a derived class to react on changes in the BLE Supervision
    @param supervision
  */
  protected void BLESupervisionChanged(shared_ptr<BleSupervision> supervision)
  {
  }

  /**
    This method can be overwritten by a derived class to react on changes in the BSA Control
    @param bsaControl
  */
  protected void BSAControlChanged(shared_ptr<BsaControl> bsaControl)
  {
  }

//--------------------------------------------------------------------------------
//@private members
//--------------------------------------------------------------------------------

  /**
    callback function to show the alertcolor on a shape
    @param userData ... user data
    @param tab ... tab on where to perform action
  */
  private static void showAlertColorOnShapeCB(anytype userData, dyn_dyn_anytype tab)
  {
    string sFrom = userData["from"];
    bool bShowAlertAsForeColor = userData["bShowAlertAsForeColor"];
    bool bShowAlertAsBackColor = userData["bShowAlertAsBackColor"];
    bool bShapeInvisibleIfNoAlert = userData["bShapeInvisibleIfNoAlert"];
    int iMinPrio = userData["minPrio"];
    int iAppNotAck = 0;
    int iAppDisappNotAck = 0;
    int iDisappNotAck = 0;
    int iAppAck = 0;

    strreplace(sFrom, "|", ",");

    dpQuery("SELECT ALERT '_alert_hdl.._alert_color', '_alert_hdl.._value', '_alert_hdl.._prior', '_alert_hdl.._last_of_all', '_alert_hdl.._act_state', '_alert_hdl.._sum' "+
            "FROM '{"+sFrom+"}' WHERE '_alert_hdl.._last_of_all' == 1 AND '_alert_hdl.._sum' == 0 AND '_alert_hdl.._prior' >= "+iMinPrio+" SORT BY 3 DESC", tab);

    setValue(userData["shape"], "foreCol", "");
    setValue(userData["shape"], "backCol", "");

    if ( dynlen(tab) < 2 )//alarm not visible
    {
      if ( bShapeInvisibleIfNoAlert )
        setValue(userData["shape"], "visible", false);

      return;
    }
    else
      setValue(userData["shape"], "visible", true);

    for(int i = 2 ; i<=dynlen(tab) ; ++i)
    {
      if(tab[i][7] == DPATTR_ALERTSTATE_APP_NOT_ACK && iAppNotAck == 0)
      {
        iAppNotAck = i;
      }
      else if(tab[i][7] == DPATTR_ALERTSTATE_APP_DISAPP_NOT_ACK && iAppDisappNotAck == 0)
      {
        iAppDisappNotAck = i;
      }
      else if(tab[i][7] == DPATTR_ALERTSTATE_DISAPP_NOT_ACK && iDisappNotAck == 0)
      {
        iDisappNotAck = i;
      }
      else if(tab[i][7] ==  DPATTR_ALERTSTATE_APP_ACK && iAppAck == 0)
      {
        iAppAck = i;
      }
    }

    string color;

    if(iAppNotAck)
    {
      color = tab[iAppNotAck][3];
    }
    else if(iAppDisappNotAck)
    {
      color = tab[iAppDisappNotAck][3];
    }
    else if(iDisappNotAck)
    {
      color = tab[iDisappNotAck][3];
    }
    else if(iAppAck)
    {
      color = tab[iAppAck][3];
    }
    else   //no alert
    {
      if ( bShapeInvisibleIfNoAlert )
        setValue(userData["shape"], "visible", false);

      return;
    }

    if ( bShowAlertAsForeColor )
      setValue(userData["shape"], "foreCol", color);
    if ( bShowAlertAsBackColor )
      setValue(userData["shape"], "backCol", color);
  }

  /**
    Callback signaling a change in the BLE Supervision
    @param supervision
  */
  private void BLESupervisionChangedCB(shared_ptr<BleSupervision> supervision)
  {
    //DebugTN(supervision.GetRawSupervision());
    ModeControlSendUpdate();

    BLESupervisionChanged(supervision);
  }

  /**
    Callback signaling a change in the BSA Control
    @param bsaControl
  */
  private void BSAControlChangedCB(shared_ptr<BsaControl> bsaControl)
  {
    assignPtr(_bsaControl, bsaControl);

    ModeServeSendUpdate();
    ModeControlSendUpdate();

    BSAControlChanged(bsaControl);
  }

  /**
    Callback signaling a change in mode_serve and/or mode_control
    @param serveDp
    @param modeServe
    @param controlDp
    @param modeControl
  */
  private void ModeServeControlCB(string serveDp,   int modeServe,
                                  string controlDp, int modeControl,
                                  string spezDp,    int spez)
  {
    _deviceModeServe = modeServe;
    _currentModeControl = modeControl;
    _currentSpez = spez;

    ModeServeSendUpdate();
    ModeControlSendUpdate();
    SpezSendUpdate();
  }

  //Steuerungsarts
  /**
    Sends an update to indicate a change in modeControl (Steuerungsart)
  */
  private void ModeControlSendUpdate()
  {
    bool modeControlAuthorized =  _workstation.IsAuthorized(_objectDpName + ".cmdio.mode_control");  //mma: check if for the mode_control dpe authorization is granted

    _modeControlAllowed = modeControlAuthorized && IsModeServeRemote();  //check if mode_serve is also remote (BLE or BSA)

    synchronized(_controlChangeListenerMutex)
    {
      for(int i=0 ; i < _modeControlChangeListener.count() ; ++i)  //mma: inform all registered listeners about the changed modeControl
      {
        shape listener = _modeControlChangeListener.at(i);
        listener.ModeControlChangedCB(_currentModeControl, _modeControlAllowed);
      }
    }
  }

  //Bedienart
  /**
    Sends an update to indicate a change in modeServe (Bedienart)
  */
  private void ModeServeSendUpdate()
  {
    _currentModeServe = EvaluateOverallModeServe();

    synchronized(_serveChangeListenerMutex)
    {
      for(int i=0 ; i < _modeServeChangeListener.count() ; ++i)  //mma: inform all registered listeners about the changed modeServe
      {
        shape listener = _modeServeChangeListener.at(i);
        listener.ModeServeChangedCB(_currentModeServe);
      }
    }
  }

  /**
    Calculates the overall modeServe based on the deviceMode serve and the BSAControl
    @return
  */
  private BsaModeServe EvaluateOverallModeServe()
  {
    BsaModeServe devModeServe;
    BsaModeServe overallModeServe;

    try
    {
      devModeServe = _deviceModeServe;
    }
    catch  //mma: in case the int value does convert to the emun, set it to BLE
    {
      devModeServe = BsaModeServe::BLE;
    }

    if(devModeServe == BsaModeServe::BLE && _bsaControl.IsBsaControl())
    {
      overallModeServe = BsaModeServe::BSA;
    }
    else
    {
      overallModeServe = devModeServe;
    }

    return overallModeServe;
  }

  /**
    Sends an update to indicate a change in spez
  */
  private void SpezSendUpdate()
  {
    synchronized(_spezChangeListenerMutex)
    {
      for(int i=0 ; i < _spezChangeListener.count() ; ++i)  //mma: inform all registered listeners about the changed spez
      {
        shape listener = _spezChangeListener.at(i);
        listener.SpezChangedCB(_currentSpez);
      }
    }
  }

//--------------------------------------------------------------------------------
//@protected data
//--------------------------------------------------------------------------------
  protected shared_ptr<BsaWorkstation> _workstation;

//--------------------------------------------------------------------------------
//@private data
//--------------------------------------------------------------------------------
  private string _objectDpName;

  private shared_ptr<BsaControl> _bsaControl;

  private BsaModeControl _currentModeControl;
  private bool _modeControlAllowed;
  private vector<shape> _modeControlChangeListener;
  private bool _controlChangeListenerMutex;

  private BsaModeServe _currentModeServe;
  private int _deviceModeServe;
  private vector<shape> _modeServeChangeListener;
  private bool _serveChangeListenerMutex;

  private BsaSpez _currentSpez;
  private vector<shape> _spezChangeListener;
  private bool _spezChangeListenerMutex;
};
