// $License: NOLICENSE
//--------------------------------------------------------------------------------
/**
  @file $relPath
  @copyright $copyright
  @author mmarchhart
*/

//--------------------------------------------------------------------------------
// Libraries used (#uses)
#uses "classes/bsa/BsaPanelNavigation"

//--------------------------------------------------------------------------------
// Variables and Constants

//--------------------------------------------------------------------------------
/**
*/

enum BsaScreenType
{
  BSA_HMI,
  BLE_INTEGRATED_HMI
};

struct NavigationItem
{
  NavigationItem(const string &cnsPanelPath, bool historyNavigation)
  {
    CnsPanelPath = cnsPanelPath;
    HistoryNavigation = historyNavigation;
  }

  string CnsPanelPath;
  bool HistoryNavigation;
};

struct HierarchyNavOption
{
  string LeftCnsPanelPath;
  string RightCnsPanelPath;
  string UpCnsPanelPath;
};

class BsaScreen
{
//--------------------------------------------------------------------------------
//@public members
//--------------------------------------------------------------------------------

  //------------------------------------------------------------------------------
  /** The Default Constructor.
  */
  public BsaScreen(int screenNumber, BsaScreenType screenType = BsaScreenType::BSA_HMI)
  {
    _screenNumber = screenNumber;
    _screenType = screenType;

    _baseModuleName = "baseModule_" + screenNumber;
    _screenPanelName = "mainPanel_" + screenNumber;
    _mainModuleName  = "mainModule_" + screenNumber;

    _menuModuleName    = "menuModule_"  + screenNumber;
    _navModuleName     = "navModule_"  + screenNumber;
    _pictureModuleName = "pictureModule_"  + screenNumber;
    _alertModuleName   = "alertModule_"  + screenNumber;
    _functionModuleName = "functionModule_"  + screenNumber;

    _navWorkerThread = -1;
  }

  public void ShowScreen(string cnsStartPanelPath="")
  {
    string basePanelFile = "";

    switch(_screenType)
    {
      case BsaScreenType::BSA_HMI :
        basePanelFile = "bsa/bsaBaseBSA_HMI.xml";
      break;

      case BsaScreenType::BLE_INTEGRATED_HMI :
        basePanelFile = "bsa/bsaBaseBLE_INTEGRATED_HMI.xml";
      break;

      default:
        throwError(makeError("bsa_framework", PRIO_SEVERE, ERR_IMPL, 100, _screenType));
        return;
    }

    ModuleOn(_baseModuleName);
    titleBar(false, _baseModuleName);
    moduleMaximize(_baseModuleName);

    RootPanelOnModule(basePanelFile, _screenPanelName, _baseModuleName, makeDynString());

    while(!isPanelOpen(_screenPanelName, _baseModuleName))  //mma: wait till the panel has loaded
    {
      delay(0,500);
    }

    if(cns_nodeExists(cnsStartPanelPath))
    {
      _navHierarchyTopMostCnsPanelPath = cnsStartPanelPath;
      NavigateToImpl(new NavigationItem(cnsStartPanelPath, false));
    }
  }

  /**
    Request to navigate to a different panel via the cns path
    @param cnsPanelPath
  */
  public void NavigateTo(const string &cnsPanelPath)
  {
    EnqueueNavRequest(new NavigationItem(cnsPanelPath, false));
  }

  /**
    Request to go one step back in the panel history
  */
  public void NavigateHistoryBack()
  {
    if(_navHistoryPtr > 0)  //mma: only go back in history if not already at the oldest entry
    {
      --_navHistoryPtr;
      EnqueueNavRequest(new NavigationItem(_navigationHistory.at(_navHistoryPtr), true));
    }
  }

  /**
    Request to go one step forward in the panel history
  */
  public void NavigateHistoryForward()
  {
    if(_navHistoryPtr < _navigationHistory.count() -1)  //mma: only go forward in history if not already at the newest entry
    {
      ++_navHistoryPtr;
      EnqueueNavRequest(new NavigationItem(_navigationHistory.at(_navHistoryPtr), true));
    }
  }

  /**
    Request to navigate one step up in the panel hirarchy
  */
  public void NavigateHirarchyUp()
  {
    HierarchyNavOption navOption = GetCurrentHierarchyNavOption();

    if(navOption.UpCnsPanelPath != "")  //mma: if navigation one level up is possible
    {
      EnqueueNavRequest(new NavigationItem(navOption.UpCnsPanelPath, false));
    }
  }

  /**
    Request to navigate one step to the left in the panel hirarchy
  */
  public void NavigateHirarchyLeft()
  {
    HierarchyNavOption navOption = GetCurrentHierarchyNavOption();

    if(navOption.LeftCnsPanelPath != "")  //mma: if navigation to the left is possible
    {
      EnqueueNavRequest(new NavigationItem(navOption.LeftCnsPanelPath, false));
    }
  }

  /**
    Request to navigate one step to the right in the panel hirarchy
  */
  public void NavigateHirarchyRight()
  {
    HierarchyNavOption navOption = GetCurrentHierarchyNavOption();

    if(navOption.RightCnsPanelPath != "")  //mma: if navigation to the right is possible
    {
      EnqueueNavRequest(new NavigationItem(navOption.RightCnsPanelPath, false));
    }
  }

  /**
    Starts the naviation worker thread.
    This must be called once in the init function of the basePanel so the thread runs in this context
  */
  public void StartNavigationWorker()
  {
    if(_navWorkerThread == -1)  //mma: start the thread only if not already started
    {
      throwError(makeError("bsa_framework", PRIO_INFO, ERR_IMPL, 103));

      _navWorkerThread = startThread(this, NavigationWorker);
    }
    else
    {
      throwError(makeError("bsa_framework", PRIO_WARNING, ERR_IMPL, 104));
    }
  }

  //Getters
  public string GetBaseModuleName()
  {
    return _baseModuleName;
  }

  public string GetMainModuleName()
  {
    return _mainModuleName;
  }

  public string GetMenuModuleName()
  {
    return _menuModuleName;
  }

  public string GetNavModuleName()
  {
    return _navModuleName;
  }

  public string GetPictureModuleName()
  {
    return _pictureModuleName;
  }

  public string GetAlertModuleName()
  {
    return _alertModuleName;
  }

  public string GetFunctionModuleName()
  {
    return _functionModuleName;
  }

  public BsaScreenType GetScreenType()
  {
    return _screenType;
  }

  public int GetScreenNumber()
  {
    return _screenNumber;
  }

  /**
    Registers a new navigation change listener
    The reference panel must implement public void NavigationChangedCB(const string &newCnsPanelPath) to be a valid listener
    @param listener
  */
  public void RegisterNavigationChangeListener(shape listener)
  {
    synchronized(_changeListenerMutex)
    {
      _navChangeListener.append(listener);

      listener.NavigationChangedCB(_currentCnsPanelPath);  //mma: inital inform the listener about the current shown panel
    }
  }

  /**
    Removes a navigation change listener
    @param listener
  */
  public void RemoveNavigationChangeListener(shape listener)
  {
    synchronized(_changeListenerMutex)
    {
      int index = _navChangeListener.indexOf(listener);

      if(index > -1)  //mma: listener found in the list
      {
        _navChangeListener.removeAt(index);
      }
    }
  }

  /**
    Registers a new navigation history change listener
    The reference panel must implement public void NavHistoryChangedCB(int currentPosition, int totalHistoryEntries) to be a valid listener
    @param listener
  */
  public void RegisterNavHistoryChangeListener(shape listener)
  {
    synchronized(_historyChangeListenerMutex)
    {
      _navHistoryChangeListener.append(listener);

      listener.NavHistoryChangedCB(_navHistoryPtr+1, _navigationHistory.count());  //mma: inital inform the listener about the current history status
    }
  }

  /**
    Removes a navigation history change listener
    @param listener
  */
  public void RemoveNavHistoryChangeListener(shape listener)
  {
    synchronized(_historyChangeListenerMutex)
    {
      int index = _navHistoryChangeListener.indexOf(listener);

      if(index > -1)  //mma: listener found in the list
      {
        _navHistoryChangeListener.removeAt(index);
      }
    }
  }

  /**
    Registers a new navigation hierarchy change listener
    The reference panel must implement public void NavHierarchyChangedCB(const HierarchyNavOption &hierarchyNavOption) to be a valid listener
    @param listener
  */
  public void RegisterHierarchyChangeListener(shape listener)
  {
    synchronized(_hierarchyChangeListenerMutex)
    {
      _navHierarchyChangeListener.append(listener);

      HierarchyNavOption navOption = GetCurrentHierarchyNavOption();

      listener.NavHierarchyChangedCB(navOption);  //mma: inital inform the listener about the current hierarchy status
    }
  }

  /**
    Removes a navigation hierarchy change listener
    @param listener
  */
  public void RemoveHierarchyChangeListener(shape listener)
  {
    synchronized(_hierarchyChangeListenerMutex)
    {
      int index = _navHierarchyChangeListener.indexOf(listener);

      if(index > -1)  //mma: listener found in the list
      {
        _navHierarchyChangeListener.removeAt(index);
      }
    }
  }

//--------------------------------------------------------------------------------
//@public static members
//--------------------------------------------------------------------------------
  public static BsaPanelLoadingType IntToLoadingType(int type)
  {
    BsaPanelLoadingType loadingType;

    switch(type)
    {
      case 0 :
        loadingType = BsaPanelLoadingType::STANDARD;
        break;

      case 1 :
        loadingType = BsaPanelLoadingType::DETAIL;
        break;

      case 2 :
        loadingType = BsaPanelLoadingType::LOAD_FIRST_DETAIL;
        break;

      default :
        loadingType = BsaPanelLoadingType::STANDARD;
        break;
    }
    return loadingType;
  }

//--------------------------------------------------------------------------------
//@protected members
//--------------------------------------------------------------------------------

  //--------------------------------------------------------------------------------
//@private members
//--------------------------------------------------------------------------------
  private void EnqueueNavRequest(shared_ptr<NavigationItem> navItem)
  {
    if(_navWorkerThread != -1)
    {
      synchronized(_nextNavRequestMutex)
      {
        _nextNavigationRequest.append(navItem);
      }
    }
    else
    {
      throwError(makeError("bsa_framework", PRIO_SEVERE, ERR_IMPL, 101));
    }
  }

  private void NavigationWorker()
  {
    while(true)
    {
      synchronized(_nextNavRequestMutex)
      {
        if(_nextNavigationRequest.count() > 0)  //mma: check if there is a new navigation request in the queue
        {
          NavigateToImpl(_nextNavigationRequest.takeFirst());
        }
      }

      delay(0, 50);
    }

    _navWorkerThread = -1;
  }

  private void NavigateToImpl(shared_ptr<NavigationItem> navItem)
  {
    string cnsPanelPath = navItem.CnsPanelPath;

    BsaPanelNavDetail navDetail = BsaPanelNavigation::getNavDetails(cnsPanelPath);

    try
    {
      if(navDetail.LoadingType == BsaPanelLoadingType::DETAIL)  //mma: in case of a detailed panel, check first if the correct main panel is loaded
      {
        string parentCnsPath;
        cnsGetParent(cnsPanelPath, parentCnsPath);

        if(_currentCnsMainModulePanelPath != parentCnsPath)  //mma: in case the current main panel is not the main panel requiered for the requested detail panel, load first the correct main panel
        {
          BsaPanelNavDetail parentNavDetail = BsaPanelNavigation::getNavDetails(parentCnsPath);
          OpenPanel(parentNavDetail);
        }

        OpenPanel(navDetail);  //mma: now open the detail panel
      }
      else if(navDetail.LoadingType == BsaPanelLoadingType::LOAD_FIRST_DETAIL)  //mma: in this case, open the main panel and then the first detail panel (first child in CNS path)
      {
        OpenPanel(navDetail);

        dyn_string childCnsPaths;
        cnsGetChildren(cnsPanelPath, childCnsPaths);

        if(childCnsPaths.count() > 0)  //mma: if there are child panels defined in the cns hirachy
        {
          BsaPanelNavDetail childNavDetail = BsaPanelNavigation::getNavDetails(childCnsPaths.at(0));  //mma: get the panel navigation details of the first child

          OpenPanel(childNavDetail);

          navItem.CnsPanelPath = childNavDetail.CNSPath;  //mma: set the cnsPanelPath in the navItem to the child panel so the SetCurrentCnsPanelPath get the correct one
        }
      }
      else
      {
        OpenPanel(navDetail);
      }

      SetCurrentCnsPanelPath(navItem);
    }
    catch
    {
      throwError(getLastException());
      DebugTN("could not open panel: " + cnsPanelPath);  //TODO: Open here poupu with the content of getLastError();

    }
  }

  private void OpenPanel(const BsaPanelNavDetail &panelNavDetail)
  {
    if(isfile(getPath(PANELS_REL_PATH, panelNavDetail.PanelFile)))  //mma: check if specified file exists
    {
      string moduleName;

      if(panelNavDetail.LoadingType == BsaPanelLoadingType::DETAIL)  //mma: in case it is a detail panel, load it into the picture module, otherwise exchange the content of the whole main module
      {
        moduleName = _pictureModuleName;
      }
      else
      {
        moduleName = _mainModuleName;
        _currentCnsMainModulePanelPath = panelNavDetail.CNSPath;
      }

      SetTitles(panelNavDetail.LoadingType, panelNavDetail.Description);

      string panelName = delExt(panelNavDetail.PanelFile);
      strreplace(panelName, "/", "_");
      strreplace(panelName, "\\", "_");

      RootPanelOnModule(panelNavDetail.PanelFile, panelName, moduleName, makeDynString());

      while(!isPanelOpen(panelName, moduleName))  //mma: wait till the panel has loaded
      {
        delay(0,50);
      }
    }
    else
    {
      throw(makeError("bsa_framework", PRIO_SEVERE, ERR_PARAM, 102, "cnsPath: " + panelNavDetail.CNSPath + " file: " + panelNavDetail.PanelFile));
    }
  }

  private void SetTitles(BsaPanelLoadingType loadingType, const string &description)
  {
    if(_screenType == BsaScreenType::BSA_HMI)  //mma: set titles only if screen started as BSA_HMI (in BLE_INTEGRATED_HMI the text field shapes do not exist)
    {
      if(loadingType == BsaPanelLoadingType::DETAIL) //mma: in case of a detail panel, only modify the title 3
      {
        setValue(_baseModuleName + "." + _screenPanelName + ":txtTitle3", "text", description);
      }
      else if(loadingType == BsaPanelLoadingType::LOAD_FIRST_DETAIL)
      {
        setMultiValue(_baseModuleName + "." + _screenPanelName + ":txtTitle2", "text", description,
                      _baseModuleName + "." + _screenPanelName + ":txtTitle3", "text", "");
      }
      else
      {
        setMultiValue(_baseModuleName + "." + _screenPanelName + ":txtTitle2", "text", description,
                      _baseModuleName + "." + _screenPanelName + ":txtTitle3", "text", description);
      }
    }
  }

  private void SetCurrentCnsPanelPath(shared_ptr<NavigationItem> navItem)
  {
    synchronized(_changeListenerMutex)
    {
      _currentCnsPanelPath = navItem.CnsPanelPath;

      for(int i=0 ; i < _navChangeListener.count() ; ++i)  //mma: inform all registered listeners about the changed panel
      {
        shape listener = _navChangeListener.at(i);
        listener.NavigationChangedCB(_currentCnsPanelPath);
      }
    }

    if(!navItem.HistoryNavigation)  //mma: if it was not due to a history navigation request, append it to the history list
    {
      _navigationHistory.append(navItem.CnsPanelPath);
      _navHistoryPtr = _navigationHistory.count() -1;  //mma: set the history navigation ptr to the last entry
    }

    synchronized(_historyChangeListenerMutex)
    {
      for(int i=0 ; i < _navHistoryChangeListener.count() ; ++i)  //mma: inform all registered history changed listeners about the change
      {
        shape listener = _navHistoryChangeListener.at(i);
        listener.NavHistoryChangedCB(_navHistoryPtr+1, _navigationHistory.count());
      }
    }

    HierarchyNavOption navOption = GetCurrentHierarchyNavOption();

    synchronized(_hierarchyChangeListenerMutex)
    {
      for(int i=0 ; i < _navHierarchyChangeListener.count() ; ++i)  //mma: inform all registered hierarchy changed listeners about the change
      {
        shape listener = _navHierarchyChangeListener.at(i);
        listener.NavHierarchyChangedCB(navOption);
      }
    }
  }

  /**
    Returns the current hierarchy navigation possibilities (left, right and up)
    @return
  */
  private HierarchyNavOption GetCurrentHierarchyNavOption()
  {
    HierarchyNavOption navOption;

    if(_currentCnsPanelPath != "")
    {
      if(BsaPanelNavigation::GetLoadingType(_currentCnsPanelPath) == BsaPanelLoadingType::DETAIL)  //mma: navigation to left and right is only possible in chase the current shown panel is a detail panel
      {
        navOption.LeftCnsPanelPath = BsaPanelNavigation::GetLeftCNSPanelPath(_currentCnsPanelPath);  //mma: if there is a panel then navigation to the left is possible
        navOption.RightCnsPanelPath = BsaPanelNavigation::GetRightCNSPanelPath(_currentCnsPanelPath);  //mma: if there is a panel then navigation to the right is possible
      }

      navOption.UpCnsPanelPath = GetNextUpCNSPanelPath(_currentCnsPanelPath);  //mma: if there is a panel then navigation up is possible
    }
    return navOption;
  }

  /**
    Searches for the next upper panel in the panel topology that is a standard panel (BsaPanelLoadingType::STANDARD) and returns the respective cnsPanelPath.
    In case no further upper panel exists (or the topmost panel is already reached) an empty string is returned.
    @param cnsPanelPath
    @return
  */
  private string GetNextUpCNSPanelPath(string cnsPanelPath)
  {
    string cnsParentPath;

    if(cnsPanelPath == _navHierarchyTopMostCnsPanelPath)  //mma: in case we are already at the topmost panel
    {
      return "";
    }
    else if(cnsGetParent(cnsPanelPath, cnsParentPath))  //mma: get the parent cns path
    {
      if(BsaPanelNavigation::GetLoadingType(cnsParentPath) != BsaPanelLoadingType::STANDARD)  //mma: in case the parent panel is a non-standard panel we check further
      {
        if(cnsParentPath != _navHierarchyTopMostCnsPanelPath)  //mma: if the non-standart parent panel panel is not the topmost panel, keep searching (recursivly) for the next upper panel till it is a standard one or the topmost one
        {
          return GetNextUpCNSPanelPath(cnsParentPath);
        }
        else return "";
      }
      else
      {
         return cnsParentPath;
      }
    }
    else return "";

  }
//--------------------------------------------------------------------------------
//@private data
//--------------------------------------------------------------------------------
  private int _screenNumber;
  private BsaScreenType _screenType;

  private string _baseModuleName;
  private string _mainModuleName;
  private string _screenPanelName;

  private string _menuModuleName;
  private string _navModuleName;
  private string _pictureModuleName;
  private string _alertModuleName;
  private string _functionModuleName;

  private string _currentCnsPanelPath;
  private string _currentCnsMainModulePanelPath;
  private vector<shape> _navChangeListener;
  private bool _changeListenerMutex;

  private vector< shared_ptr<NavigationItem> > _nextNavigationRequest;
  private bool _nextNavRequestMutex;
  private int _navWorkerThread;

  private vector<string> _navigationHistory;
  private uint _navHistoryPtr;
  private vector<shape> _navHistoryChangeListener;
  private bool _historyChangeListenerMutex;

  private string _navHierarchyTopMostCnsPanelPath;
  private vector<shape> _navHierarchyChangeListener;
  private bool _hierarchyChangeListenerMutex;
};
