// $License: NOLICENSE
//--------------------------------------------------------------------------------
/**
  @file $relPath
  @copyright $copyright
  @author mmarchhart
*/

//--------------------------------------------------------------------------------
// Libraries used (#uses)
#uses "classes/bsa/BsaScreen"
#uses "classes/bsa/BsaControl"
#uses "classes/bsa/BsaDpGroupCache"
#uses "classes/bsa/BsaUser"
#uses "classes/bsa/BleSupervision"
#uses "classes/bsa/BleSupervisionUserGroup"

//--------------------------------------------------------------------------------
// Variables and Constants

//--------------------------------------------------------------------------------
/**
*/
class BsaWorkstation
{
//--------------------------------------------------------------------------------
//@public members
//--------------------------------------------------------------------------------
  /**
    Gets the sole instance of the workstation (singleton)
    @return
  */
  public static shared_ptr<BsaWorkstation> GetInstance()
  {
    if(_instance == nullptr)
    {
      _instance = new BsaWorkstation();
    }

    return _instance;
  }

  /**
    Initializes the workstation
  */
  public void InitBsaWorkstation(BsaScreenType screenType, shape workstationShape, BsaUser bsaUser)
  {
    DebugTN(bsaUser);
    _screenType = screenType;
    _workstationShape = workstationShape;
    _workstationModule = myModuleName();

    _currentUser = bsaUser;

    //create and init the dpGroup cache
    _dpGroupCache = new BsaDpGroupCache();
    _dpGroupCache.InitDPGroupCache();

    //mma: connect to the bsaControl DPE
    dpConnect(this, this.BsaControlCB, "_BsaCentral.bsaControl");

    //mma: init the dpConnection to get the "BLE Führung" information
    dyn_string systemMainDP;
    dpGet("_BsaBaseConfig.systemDP.systemMain", systemMainDP);

    if(systemMainDP.count() > 0)
    {
      dpConnect(this, this.BleSupervisionCB, systemMainDP.at(0) + ".cmdio.supervision");
    }
  }

  /**
    Starts a new screen instance of the given screentType
    @param screenType
    @param cnsStartPanelPath
    @return
  */
  public int StartBsaScreen(string cnsStartPanelPath="")
  {
    shared_ptr<BsaScreen> screen = new BsaScreen(_screens.count() + 1, _screenType);
    _screens.append(screen);

    screen.ShowScreen(cnsStartPanelPath);

    return _screens.count();
  }

  /**
    Closes the screen instance the calling module belongs to
  */
  public void CloseMyScreen()
  {
    int screenNumber = GetMyScreenNumber();

    CloseBsaScreen(screenNumber);
  }

  /**
    Closes the screen instance via the given screen number
    @param screenNumber
  */
  public void CloseBsaScreen(int screenNumber)
  {
    shared_ptr<BsaScreen> screen = GetScreen(screenNumber);

    if(screen != nullptr)
    {
      _screens.removeAt(_screens.indexOf(screen));  //mma: first remove the screen from the active list

      if(_screens.count() == 0)  //mma: if it was the last screen, close also the workstation module
      {
        if(screen.GetScreenType() == BsaScreenType::BSA_HMI)  //mma: if in addition it is a BSA_HMI, also call the ReleaseBSAControl()
        {
          ReleaseBSAControl();
        }

        ModuleOff(_workstationModule);
      }

      //finally close the screen module
      ModuleOff(screen.GetBaseModuleName());
    }
  }

  /**
    Returns the screen number the calling module belongs to
    @return
  */
  public int GetMyScreenNumber()
  {
    dyn_string split = strsplit(myModuleName(),"_");

    if(split.count() == 2)
    {
      return split.at(1);
    }
    else return 0;
  }

  /**
    Returns the screen instance the calling module belongs to
    In case the moudule does not belong to a BsaScreen, a nullptr is returned
    @return
  */
  public shared_ptr<BsaScreen> GetMyScreen()
  {
    int screenNumber = GetMyScreenNumber();

    return GetScreen(screenNumber);
  }

  /**
    Returns the screen instance via the given screen number
    In case no matching BsaScreen exists, a nullptr is returned
    @return
  */
  public shared_ptr<BsaScreen> GetScreen(int screenNumber)
  {
    for(int i=0 ; i < _screens.count() ; ++i)
    {
      shared_ptr<BsaScreen> screen = _screens.at(i);

      if(screen.GetScreenNumber() == screenNumber)
      {
        return screen;
      }
    }

    //mma: in case no screen was found, return nullptr
    return nullptr;
  }

  /**
    Navigates to the given cnsPanelPath on the screen the calling module belongs to
    @param cnsPanelPath
  */
  public void NavigateTo(string cnsPanelPath)
  {
    shared_ptr<BsaScreen> screen = GetMyScreen();

    if(screen != nullptr)
    {
      screen.NavigateTo(cnsPanelPath);
    }
  }

  /**
    Changes the state of the BSA Control using the following pattern:
    If no one has currently BSA Control -> assigne BSA Control to this user
    If another user has currently BSA Control -> remove the control from the other user and assigne it to this user
    If this user has currently BSA Control -> release BSA Control
  */
  public void SwitchBSAControl()
  {
    string currentUser = _currentUser.GetUserName();
    shared_ptr<BsaControl> currentBsaControl = GetCurrentBsaControl();

    if(!currentBsaControl.IsBsaControl())  //mma: if currently not in BSA control, then aquire it for the current user
    {
      BsaControl newBsaControl = BsaControl(currentUser, myUiNumber());
      dpSetWait("_BsaCentral.bsaControl", newBsaControl.ToString());
    }
    else   //mma: if currently under BSA control
    {
      if(currentBsaControl.GetUserName() == currentUser)  //mma: if it is aquired by the current user then release the control
      {
        BsaControl emptyBsaControl;
        dpSetWait("_BsaCentral.bsaControl", emptyBsaControl.ToString());
      }
      else   //mma: if currently aquired by another user, swtich over to this user
      {
        BsaControl newBsaControl = BsaControl(currentUser, myUiNumber());
        dpSetWait("_BsaCentral.bsaControl", newBsaControl.ToString());
      }
    }
  }

  /**
    Releases the BSA Control if currently assinged to this user
  */
  public void ReleaseBSAControl()
  {
    string currentUser = _currentUser.GetUserName();
    shared_ptr<BsaControl> currentBsaControl = GetCurrentBsaControl();

    if(currentBsaControl.IsBsaControl() && currentBsaControl.GetUserName() == currentUser)  //mma: if currently under BSA control and aquired by the current user then release the control
    {
      BsaControl emptyBsaControl;
      dpSetWait("_BsaCentral.bsaControl", emptyBsaControl.ToString());
    }
  }

  /**
    Returns the current BSA Control
    @return
  */
  public shared_ptr<BsaControl> GetCurrentBsaControl()
  {
    return _currentBsaControl;
  }

  /**
    Retursn the current BLE Supervision
    @return
  */
  public shared_ptr<BleSupervision> GetCurrentBleSupervision()
  {
    return _currentBleSupervision;
  }

  /**
    Returns the current user name
    @return
  */
  public BsaUser GetCurrentUser()
  {
    return _currentUser;
  }

  /**
    Searches for the supervision area the given DPE belongs to. In case of no match, an empty string is returned.
    HINT: The assignement is implemented using DPGroups. So for each supervision area a corresponding DPGroup must exist and the DPE assigned to exactly one of those groups
    @param dpe in the format DP.dpe
    @return name of the supervision area to which the given dpe belongs to
  */
  public string FindDPESupervisionArea(const string &dpe)
  {
    return _dpGroupCache.FindDPEGroup(getSystemName() + dpe);
  }

  /**
    Returns the workstation shape.
    This shape can be used to connect to "global" UI events
    current events:
    - BSAControlChanged(shared_ptr<BsaControl> bsaControl)
    - BLESupervisionChanged(shared_ptr<BleSupervision> supervision)

    @return
  */
  public shape GetWorkstationShape()
  {
    return _workstationShape;
  }

  /**
    Checks if currently authorization for the given DPE is granted.
    BLE_Integrated:
    It depends on the supervision area the DPE belongs to, the bleSupervion group the current user belongs to and if this group has currently authorization in this supervions area
    In addition it depends on the BSAControl status (BSAControl overrouls!!)
    BSA:
    It depends if the current user has BSA control or not

    @param dpe
    @return
  */
  public bool IsAuthorized(const string &dpe)
  {
    if(_screenType == BsaScreenType::BSA_HMI)  //mma: in case of BSA_HMI
    {
      if(_currentBsaControl.IsBsaControl() && _currentBsaControl.GetUserName() == _currentUser.GetUserName())  //mma: if currently BSA control claimbed by the current user
      {
        return true;
      }
    }
    else if(_screenType == BsaScreenType::BLE_INTEGRATED_HMI && !_currentBsaControl.IsBsaControl())  //mma: in case of BLE_INTEGRATED_HMI and NOT BSA control
    {
      string supervisionArea = FindDPESupervisionArea(dpe);

      return _currentBleSupervision.IsAuthorized(_currentUser.GetUserGroup(), supervisionArea);  //mma: check if the current user (via its group assignment) has authorization in the supervision area the dpe belongs to
    }

    return false;
  }
//--------------------------------------------------------------------------------
//@protected members
//--------------------------------------------------------------------------------

//--------------------------------------------------------------------------------
//@private members
//--------------------------------------------------------------------------------

  private BsaWorkstation()
  {
    _workstationModule = "";
  }

  ~BsaWorkstation()
  {
    DebugTN("Destroying BsaWorkstation");
  }

  /**
    This function gets called every time the BSAControl changes
    @param dp1
    @param bsaControlString
  */
  private void BsaControlCB(string dp1, string bsaControlString)
  {

    assignPtr(_currentBsaControl, BsaControl::FromString(bsaControlString));

    _workstationShape.TriggerBSAControlChangedEvent(_currentBsaControl);
  }

  /**
    This function gets called every time the "Führungsstring" sent from the BLE changes
    @param dp1
    @param supervisionString
  */
  private void BleSupervisionCB(string dp1, string supervisionString)
  {
    DebugTN("BLESupervision updated");


    if(_currentBleSupervision == nullptr || _currentBleSupervision.GetRawSupervision() != supervisionString)  //mma: only emit a change trigger is there is really a change in the supervision (BLE updates the string every 20 seconds!!)
    {
      assignPtr(_currentBleSupervision, new BleSupervision(supervisionString));

      //DebugTN(_currentBleSupervision);

      _workstationShape.TriggerBLESupervisionChangedEvent(_currentBleSupervision);
    }

    /*to emit the event via the intermediate shape is only a workarround since uiConnect() does not allow
        to connect to object member methods when connection to a global event (only to static methods but not to instance methods), but it is possible in case of connection to a event of a specific shape
    */
    //triggerGlobalEvent("BLESupervisionChangedEvent", supervisionString);
  }
//--------------------------------------------------------------------------------
//@private data
//--------------------------------------------------------------------------------
  private static shared_ptr<BsaWorkstation> _instance = nullptr;

  private vector< shared_ptr<BsaScreen> > _screens;
  private BsaScreenType _screenType;
  private string _workstationModule;
  private shape  _workstationShape;

  private shared_ptr<BsaDpGroupCache> _dpGroupCache = nullptr;

  private BsaUser _currentUser = BsaUser("", BleSupervisionUserGroup());
  private shared_ptr<BsaControl>      _currentBsaControl = nullptr;
  private shared_ptr<BleSupervision>  _currentBleSupervision = nullptr;

};
