// $License: NOLICENSE
//--------------------------------------------------------------------------------
/**
  @file $relPath
  @copyright $copyright
  @author mmarchhart
*/

//--------------------------------------------------------------------------------
// Libraries used (#uses)
#uses "classes/tcp/TcpConnection"
#uses "classes/tcp/ITcpDataReceiver"

#uses "classes/sode/commands/BaseCommand"

#uses "classes/sode/replies/BaseReply"
#uses "classes/sode/replies/ReplyParser"

#uses "classes/sode/ISodeDataReceiver"

//--------------------------------------------------------------------------------
// Variables and Constants

//--------------------------------------------------------------------------------
/**
*/
class SodeMessageHandler : ITcpDataReceiver
{
//--------------------------------------------------------------------------------
//@public members
//--------------------------------------------------------------------------------
  /**
    Ctor for sode message handler
    Use the SodeMessageHandlerFactory to create an instance!!
    This is used by the SodeMessageHandlerFactory (should not be exposed public but Ctrl++ does not know friend classes)
  */
  public SodeMessageHandler(const string &host, int port)
  {
    _autoReconnect = false;
    _disposed = false;
    _connection = new TcpConnection(host, port);

    _waitForReply = false;
    _reply = nullptr;

    _dataReceiver = nullptr;
  }

  /**
    Connects to SODE
    @return
  */
  public int Connect()
  {

    if(!_disposed)
    {
      if(!_connection.IsConnected())
      {
        _autoReconnect = true;  //mma: this indicates the CheckConnection() that the connection was established and reconnection attempts should be executed in case the connection was closed from server side
        return _connection.Connect();
      }
      else
      {
        throwError(makeError("sode", PRIO_WARNING, ERR_IMPL, 3));
        return 0;
      }
    }
    else
    {
      throwError(makeError("sode", PRIO_SEVERE, ERR_IMPL, 4));
      return -1;
    }
  }

  /**
    Disconnects from SODE
  */
  public void Disconnect()
  {
    _autoReconnect = false;
    _connection.Disconnect();
  }

  /**
    Returns the information if currenlty connected to the sode
    @return
  */
  public bool IsConnected()
  {
    return _connection.IsConnected();
  }

  /**
    Sets hostname or IP-Address of the SODE
    @param host
  */
  public void SetHost(const string &host)
  {
    _connection.SetHost(host);
  }

  /**
    Sets TCP port of the SODE
    @param port
  */
  public void SetPort(int port)
  {
    _connection.SetPort(port);
  }

  /**
    Sends a command to SODE and wait for a reply from SODE.
    In case of a timeout the reply status is set to ERROR with the message indicating that a timeout happened
    @param cmd Command to be sent to SODE
    @param timeout Timeout in second the function waits for a reply from SODE
    @return Returns a reply message
  */
  public synchronized shared_ptr<BaseReply> SendCommand(shared_ptr<BaseCommand> cmd, int timeout=2)
  {
    if(!_disposed)
    {
      if(IsConnected())  //mma: send the command only if connected
      {
        string data = cmd.SerializeToXml();

        synchronized(_replyMutex)
        {
          int ret = _connection.SendData(data + "\r\n");

          if(ret == 0)  //mma: data could be sent, therefore we wait for the response
          {
            _waitForReply = true;
          }
          else //mma: some error occoured while sending the data, therefore we return immediately with an error
          {
            return GenerateReply(cmd.GetName(), ReplyStatus::ERROR, "Sending command failed");
          }
        }

        bool isTimeout = false;
        time timeoutTime = getCurrentTime() + timeout;

        while(_reply == nullptr && !isTimeout)  //mma: wait till a reply was received or timeout expired
        {
          delay(0,100);
          isTimeout = getCurrentTime() > timeoutTime;  //mma: check if timeout time has elapsed
        }

        _waitForReply = false;

        shared_ptr<BaseReply> reply = nullptr;

        if(_reply != nullptr)  //mma: reply received, check if the reply correspondes to the previously sent command
        {
          if(_reply.GetCommand() == cmd.GetName())  //reply is the correct one
          {
            reply = _reply;
          }
          else   //mma: received reply that does not correspond to the sent command
          {
            throwError(makeError("sode", PRIO_WARNING, ERR_IMPL, 6, "command: " + cmd.GetName() + " reply: " + _reply.GetCommand()));
            reply = GenerateReply(cmd.GetName(), ReplyStatus::ERROR, "SODE reply incorrect");
          }
        }
        else   //mma: a timeout occured
        {
          throwError(makeError("sode", PRIO_WARNING, ERR_IMPL, 7, "command: " + cmd.GetName()));
          reply = GenerateReply(cmd.GetName(), ReplyStatus::ERROR, "SODE didn't reply in time");
        }

        _reply = nullptr;
        return reply;
      }
      else
      {
        throwError(makeError("sode", 	PRIO_WARNING, 	ERR_SYSTEM, 8));
        return GenerateReply(cmd.GetName(), ReplyStatus::ERROR, "Sending command failed, not connected to sode");
      }
    }
    else
    {
      throwError(makeError("sode", PRIO_SEVERE, ERR_IMPL, 4));
      return GenerateReply(cmd.GetName(), ReplyStatus::ERROR, "Sending command failed");
    }
  }

  public void RegisterDataReceiver(shared_ptr<ISodeDataReceiver> receiver)
  {
    receiver.ConnectionStatusChanged(IsConnected());  //mma: initialy inform the reciever about the current connection status
    assignPtr(_dataReceiver, receiver);
  }

  //implementation of ITcpDataReceiver
  public void DataReceived(const string &data)
  {
    //DebugTN(__FUNCTION__ + ": " + data);

    try
    {
      shared_ptr<BaseReply> reply = ReplyParser::DeserializeFromXml(data);
      HandleReply(reply);
    }
    catch
    {
      throwError(getLastException());
    }

    if(_dataReceiver != nullptr) _dataReceiver.RawReplyDataReceived(data);
  }

  //implementation of ITcpDataReceiver
  public void ConnectionStatusChanged(bool connected)
  {
    DebugTN("Connection changed: " + connected);
    if(_dataReceiver != nullptr) _dataReceiver.ConnectionStatusChanged(connected);
  }


  /**
    This is periodically called by the obsever thread from the SodeMessageHandlerFactory to check the TCP connection status (should not be exposed public but Ctrl++ does not know friend classes).
    It implementes the automatic reconnection attempts in case the connection was unexpectedly closed from the server side (e.g. SODE stopped)
  */
  public void CheckConnection()
  {
    if(_autoReconnect)
    {
      if(!_connection.IsConnected())  //mma: in case the connection is not established anymore, try to reconnect
      {
        throwError(makeError("sode", PRIO_INFO, ERR_IMPL, 2, "host: " + _connection.GetHost() + " port: " + _connection.GetPort()));
        _connection.Connect();
      }
    }
  }

  /**
    Returns the underlying TcpConnection object
    This is used by the SodeMessageHandlerFactory (should not be exposed public but Ctrl++ does not know friend classes)
    @return
  */
  public shared_ptr<TcpConnection> GetConnection()
  {
    return _connection;
  }

  /**
    Disposes the SodeClient object, in case the client is connected to a SODE the connection is closed as well
    This is used by the SodeMessageHandlerFactory (should not be exposed public but Ctrl++ does not know friend classes)
    @return
  */
  public void Dispose()
  {
    Disconnect();
    _disposed = true;
  }
//--------------------------------------------------------------------------------
//@protected members
//--------------------------------------------------------------------------------

//--------------------------------------------------------------------------------
//@private members
//--------------------------------------------------------------------------------
  private void HandleReply(shared_ptr<BaseReply> reply)
  {
    synchronized(_replyMutex)
    {
      if(_waitForReply)  //mma: if the client is waiting for a reply "enqueue" it to inform the waiting thread
      {
        assignPtr(_reply, reply);
      }
      else
      {
        throwError(makeError("sode", PRIO_WARNING, ERR_IMPL, 5, "command: " + reply.GetCommand()));
      }
    }

    if(_dataReceiver != nullptr) _dataReceiver.ReplyReceived(reply);
  }

  private shared_ptr<BaseReply> GenerateReply(const string &cmdName, ReplyStatus status, const string &msg)
  {
    ReplyContent content;
    content.Command = cmdName;
    content.Status = status;
    content.Msg = msg;

    shared_ptr<BaseReply> reply = ReplyParser::DecodeReply(content);

    return reply;
  }

//--------------------------------------------------------------------------------
//@private data
//--------------------------------------------------------------------------------
  private shared_ptr<TcpConnection> _connection;
  private bool _autoReconnect;
  private bool _disposed;

  private bool _waitForReply;
  private shared_ptr<BaseReply> _reply;
  private bool _replyMutex;

  private shared_ptr<ISodeDataReceiver> _dataReceiver;
};
