// $License: NOLICENSE
//--------------------------------------------------------------------------------
/**
  @file $relPath
  @copyright $copyright
  @author mmarchhart
*/

//--------------------------------------------------------------------------------
// Libraries used (#uses)
#uses "classes/sode/replies/BaseReply"
#uses "classes/sode/replies/GetStateRply"
#uses "classes/sode/replies/GetDisplayRply"
#uses "classes/sode/replies/GetTextConfigRply"
#uses "classes/sode/replies/GetVlcConfigRply"
#uses "classes/sode/replies/SetStateRply"
#uses "classes/sode/replies/SetDisplayRply"
#uses "classes/sode/replies/SetTextConfigRply"
#uses "classes/sode/replies/SetVlcConfigRply"

//--------------------------------------------------------------------------------
// Variables and Constants

//--------------------------------------------------------------------------------
/**
*/
class ReplyParser
{
//--------------------------------------------------------------------------------
//@public members
//--------------------------------------------------------------------------------

  //------------------------------------------------------------------------------
  /** The Default Constructor.
  */
  public ReplyParser()
  {
  }

   /**
    Deserializes a xml reply and returns the reply object.
    Throws an exception in case of errors.
    @param xmlReply
  */
  public static shared_ptr<BaseReply> DeserializeFromXml(const string &xmlReply)
  {
    string errMsg;
    int errLine, errColumn;
    ReplyContent replyContent;

    int docId = xmlDocumentFromString(xmlReply, errMsg, errLine, errColumn);

    if(docId == -1)
    {
      throw(makeError("sode", PRIO_SEVERE, ERR_IMPL, 100, errMsg));
    }

    int rootId = FindeRootNode(docId);

    int rplNodeId = xmlFirstChild(docId, rootId);

    if(xmlNodeName(docId, rplNodeId) == "reply")
    {
      mapping attributes = xmlElementAttributes(docId, rplNodeId);

      if(attributes.contains("command"))
      {
        replyContent.Command = attributes.value("command");
      }
      else throw(makeError("sode", PRIO_SEVERE, ERR_IMPL, 100, "attribute \"command\" not found"));

      if(attributes.contains("status"))
      {
        if(attributes.value("status") == "OK")
        {
          replyContent.Status = ReplyStatus::OK;
        }
        else
        {
          replyContent.Status = ReplyStatus::ERROR;
        }
      }
      else throw(makeError("sode", PRIO_SEVERE, ERR_IMPL, 100, "attribute \"status\" not found"));

      if(attributes.contains("msg"))  //mma: msg is an optional attribute
      {
        replyContent.Msg = attributes.value("msg");
      }

      dyn_uint paramNodeIds;
      xmlChildNodes(docId, rplNodeId, paramNodeIds);  //mma: get all reply parameters

      for(int i=0 ; i < paramNodeIds.count() ; ++i)
      {
        string paramName = xmlNodeName(docId, paramNodeIds.at(i));
        int valueNodeId = xmlFirstChild(docId, paramNodeIds.at(i));
        string paramValue;

        if(valueNodeId != -1)
        {
          paramValue = xmlNodeValue(docId, valueNodeId);
        }

        replyContent.Parameter[paramName] = paramValue;
      }

    }
    else
    {
      throw(makeError("sode", PRIO_SEVERE, ERR_IMPL, 100, "tag <reply> not found"));
    }

    xmlCloseDocument(docId);

    return DecodeReply(replyContent);
  }

  /**
    Returns the correct reply object based on the deserialized reply content
    @param content
    @return
  */
  public static shared_ptr<BaseReply> DecodeReply(const ReplyContent &content)
  {
    shared_ptr<BaseReply> reply;

    switch(content.Command)
    {
      case "getState" :
        reply = new GetStateRply(content);
        break;

      case "getDisplay" :
        reply = new GetDisplayRply(content);
        break;

      case "getTextConfig" :
        reply = new GetTextConfigRply(content);
        break;

      case "getVlcConfig" :
        reply = new GetVlcConfigRply(content);
        break;

      case "setState" :
        reply = new SetStateRply(content);
        break;

      case "setDisplay" :
        reply = new SetDisplayRply(content);
        break;

      case "setTextConfig" :
        reply = new SetTextConfigRply(content);
        break;

      case "setVlcConfig" :
        reply = new SetVlcConfigRply(content);
        break;

      default :
        reply = new BaseReply(content);
        break;
    }

    return reply;
  }
//--------------------------------------------------------------------------------
//@protected members
//--------------------------------------------------------------------------------

//--------------------------------------------------------------------------------
//@private members
//--------------------------------------------------------------------------------
  private static int FindeRootNode(int docId)
  {
    int currentId = xmlFirstChild(docId);

    while(currentId != -1)
    {
      string nodeName = xmlNodeName(docId, currentId);

      if(nodeName == "sode")  //mma: root node found, return the node id
      {
        return currentId;
      }
      currentId = xmlNextSibling(docId, currentId);
    }

    //mma: tag <sode> not found
    throw(makeError("sode", PRIO_SEVERE, ERR_IMPL, 100, "root tag <sode> not found"));
  }

};
