// $License: NOLICENSE
//--------------------------------------------------------------------------------
/**
  @file $relPath
  @copyright $copyright
  @author mmarchhart
*/

//--------------------------------------------------------------------------------
// Libraries used (#uses)
#uses "classes/tcp/ITcpDataReceiver"

//--------------------------------------------------------------------------------
// Variables and Constants

class TcpConnection
{
  public TcpConnection(const string &host, int port)
  {
    _host = host;
    _port = port;

    _socket = -1;
    _connected = false;
  }

  ~TcpConnection()
  {
    Disconnect();
  }

  /**
    Connect to the configured host and port
    @return
  */
  public int Connect()
  {
    if(!_connected)
    {
      synchronized(_socketMutex)
      {
        _socket = tcpOpen(_host, _port, 2);

        //if(_socket != -1)  //mma: in case of error the tcpOpen returns 0 and not -1 as documented, therefore we have to check if there was an error with getLastError()
        if(dynlen(getLastError()) == 0)
        {
          SetConnectionStatus(true);
          throwError(makeError("tcpConnection", PRIO_INFO, ERR_IMPL, 1, "host: " + _host + " port: " + _port));
          startThread(this, "CheckForData");  //start the receiving thread

          return 0;
        }
        else
        {
          tcpClose(_socket);
          throwError(makeError("tcpConnection", PRIO_WARNING, ERR_PARAM, 2, "host: " + _host + " port: " + _port));
          return -1;
        }
      }
    }
    else
    {
      throwError(makeError("tcpConnection", PRIO_WARNING, ERR_IMPL, 3));
      return 0;
    }
  }

  /**
    Disconnect from the TCP Server
  */
  public void Disconnect()
  {
    if(_connected)
    {
      synchronized(_socketMutex)
      {
        SetConnectionStatus(false);  //mma: this stopps also the receiving thread that is executing CheckForData()

        tcpClose(_socket);
        throwError(makeError("tcpConnection", PRIO_INFO, ERR_IMPL, 4, "host: " + _host + " port: " + _port));
      }
    }
  }

  /**
    Returns the information if currenlty connected to the server or not
    @return
  */
  public bool IsConnected()
  {
    return _connected;
  }

  public void SetHost(const string &host)
  {
    _host = host;
  }

  public void SetPort(int port)
  {
    _port = port;
  }

  public string GetHost()
  {
    return _host;
  }

  public int GetPort()
  {
    return _port;
  }

  /**
    Send data to the connected server
    @param data
  */
  public int SendData(const string &data)
  {
    if(_connected)
    {
      int ret = tcpWrite(_socket, data);
      if(ret > 0) return 0;
      else return -1;
    }
    else
    {
      throwError(makeError("tcpConnection", PRIO_WARNING, ERR_IMPL, 5));
      return -1;
    }
  }

  /**
    Registers a data receiver to handle data sent from the connected server
    @param receiver
  */
  public void RegisterReceiver(shared_ptr<ITcpDataReceiver> receiver)
  {
    receiver.ConnectionStatusChanged(_connected);  //mma: initial update of the current connection status
    _receivers.append(receiver);
  }

//--------------------------------------------------------------------------------
//@private members
//--------------------------------------------------------------------------------
  private void CheckForData()
  {
    while(_connected)
    {
      delay(0,100);

      string data;
      int readReturn;

      synchronized(_socketMutex)
      {
        if(_connected)  //mma: check again if still connected
        {
          //DebugTN(__FUNCTION__ +" try read data ");
          readReturn = tcpRead(_socket, data, 1);
          //DebugTN(__FUNCTION__ +" finished try read data ");
        }
      }

      if(readReturn == -1)  //mma: a return of -1 indicates that the connection was closed from server side, therefore we close also from our side
      {
        throwError(makeError("tcpConnection", PRIO_WARNING, ERR_IMPL, 6));
        Disconnect();
      }
      else if(data.length() > 0)
      {
        for(int i=0 ; i < _receivers.count() ; ++i)  //mma: send the data to all registered receivers
        {
          shared_ptr<ITcpDataReceiver> receiver = _receivers.at(i);
          receiver.DataReceived(data);
        }

        //DebugTN(__FUNCTION__ +" data sent to all receivers");
      }


    }
  }

  private void SetConnectionStatus(bool connected)
  {
    _connected = connected;

    for(int i=0 ; i < _receivers.count() ; ++i)
    {
      shared_ptr<ITcpDataReceiver> receiver = _receivers.at(i);
      receiver.ConnectionStatusChanged(connected);
    }

    throwError(makeError("tcpConnection", PRIO_INFO, ERR_IMPL, 7, connected));
  }

//--------------------------------------------------------------------------------
//@private data
//--------------------------------------------------------------------------------
  private int _socket;
  private bool _connected;
  private bool _socketMutex;
  private string _host;
  private int _port;

  private vector< shared_ptr<ITcpDataReceiver> > _receivers;
};


