// $License: NOLICENSE
//--------------------------------------------------------------------------------
/**
  @file $relPath
  @copyright $copyright
  @author mmarchhart
*/

//--------------------------------------------------------------------------------
// Libraries used (#uses)
#uses "classes/sode/SodeMessageHandler"
#uses "classes/sode/SodeMessageHandlerFactory"
#uses "classes/vms/sodeObject/SodeDisplayInfo"
#uses "classes/vms/sodeObject/SodeTextConfig"
#uses "classes/vms/sodeObject/sodeInterfaceHandler/SodeObjectInterfaceDataReceiver"

#uses "classes/sode/commands/BaseCommand"
#uses "classes/sode/commands/SetStateCmd"
#uses "classes/sode/commands/GetStateCmd"
#uses "classes/sode/commands/SetDisplayCmd"
#uses "classes/sode/commands/GetDisplayCmd"
#uses "classes/sode/commands/SetTextConfigCmd"
#uses "classes/sode/commands/GetTextConfigCmd"
#uses "classes/sode/commands/SetVlcConfigCmd"
#uses "classes/sode/commands/GetVlcConfigCmd"

#uses "classes/sode/replies/BaseReply"
#uses "classes/sode/replies/GetStateRply"
#uses "classes/sode/replies/GetTextConfigRply"

//--------------------------------------------------------------------------------
// Variables and Constants

//--------------------------------------------------------------------------------
/**
*/
class SodeObjectInterface
{
//--------------------------------------------------------------------------------
//@public members
//--------------------------------------------------------------------------------

  //------------------------------------------------------------------------------
  /** The Default Constructor.
  */
  public SodeObjectInterface(const string &sodeObjectDP)
  {
    _sodeObjectDP = sodeObjectDP;

    string host;
    int port;

    dpGet(_sodeObjectDP + ".config.host", host,
          _sodeObjectDP + ".config.port", port);

    _connection = SodeMessageHandlerFactory::CreateMessageHandler(host, port);

    shared_ptr<SodeObjectInterfaceDataReceiver> dataReceiver = new SodeObjectInterfaceDataReceiver(sodeObjectDP);
    _connection.RegisterDataReceiver(dataReceiver);

  }

  ~SodeObjectInterface()
  {
    SodeMessageHandlerFactory::DisposeMessageHandler(_connection);  //mma: dispose the undterlying sode message handler
  }

  /**
    Starts the sode object interface.
    This function opens the TCP/IP connection to sode and connects to the internal dpes to forward the requests from the dpes to sode
  */
  public void Start()
  {
    _connection.Connect();

    dpConnect(this, SetStateCB, false, _sodeObjectDP + ".set.state");
    dpConnect(this, SetDisplyCB, false, _sodeObjectDP + ".set.display");
    dpConnect(this, SetTextConfigCB, false, _sodeObjectDP + ".set.textConfig");
  }

  /**
    Stops the sode object interface.
    This functions closes the TCP/IP connection to sode
  */
  public void Stop()
  {
    //dpDisconnect(this, SetStateCB, _sodeObjectDP + ".set.state");  //TODO: figure out why the disconnect fails (maybe due to the fact that this is called from a different control thread???)
    DebugTN(__FUNCTION__ + " stopping " + _sodeObjectDP);
    _connection.Disconnect();
  }

  /**
    This is called periodically to poll the status of the sode (state, display, etc...)
  */
  public void ProcessInterface()
  {
    if(_connection.IsConnected())  //mma: if the connection is established
    {
      shared_ptr<BaseCommand> getStateCmd = new GetStateCmd();
      ProcessGetStateReply(_connection.SendCommand(getStateCmd));

      shared_ptr<BaseCommand> getTextConfigCmd = new GetTextConfigCmd();
      ProcessGetTextConfigReply(_connection.SendCommand(getTextConfigCmd));
    }
  }

//--------------------------------------------------------------------------------
//@protected members
//--------------------------------------------------------------------------------

  //--------------------------------------------------------------------------------
//@private members
//--------------------------------------------------------------------------------
  /**
    The callback to forward state change requests from the dpe vie a sode command
    @param dp1
    @param state
  */
  private void SetStateCB(string dp1, bool state)
  {
    string newState;

    if(state) newState = "ACTIVE";
    else newState = "STANDBY";

    shared_ptr<BaseCommand> cmd = new SetStateCmd(newState);

    shared_ptr<BaseReply> reply = _connection.SendCommand(cmd, 4);

    if(reply.GetStatus() == ReplyStatus::OK)  //mma: if sode replied with OK, then set the .current.state to the requested value so we don't have to wait for the polling
    {
      dpSet(_sodeObjectDP + ".current.state", state);
    }
  }

  /**
    The callback to forward display change requests from the dpe via a sode command
    @param dp1
    @param display
  */
  private void SetDisplyCB(string dp1, string display)
  {
    shared_ptr<SodeDisplayInfo> displayInfo = SodeDisplayInfo::FromJson(display);

    if(displayInfo != nullptr)
    {
      shared_ptr<BaseCommand> cmd = new SetDisplayCmd(displayInfo.Text, displayInfo.AddText, displayInfo.GetCameraUrl());

      shared_ptr<BaseReply> reply = _connection.SendCommand(cmd, 4);

      if(reply.GetStatus() == ReplyStatus::OK)  //mma: if sode replied with OK, then set the .current.disply to the requested value indicating that the request executed sucessfull
      {
        dpSet(_sodeObjectDP + ".current.display", displayInfo.ToJson());
      }
      else
      {
        SodeDisplayInfo emptyInfo = SodeDisplayInfo("");
        dpSet(_sodeObjectDP + ".current.display", emptyInfo.ToJson());
      }
    }
  }

  /**
    The callback to forward text config change requests from the dpe via a sode command
    @param dp1
    @param textConfig
  */
  private void SetTextConfigCB(string dp1, string textConfigJson)
  {
    shared_ptr<SodeTextConfig> textConfig = SodeTextConfig::FromJson(textConfigJson);

    if(textConfig != nullptr)
    {
      shared_ptr<BaseCommand> cmd = new SetTextConfigCmd(textConfig.PositionAsString(), textConfig.Size, textConfig.Color);

      shared_ptr<BaseReply> reply = _connection.SendCommand(cmd);

      if(reply.GetStatus() == ReplyStatus::OK)  //mma: if sode replied with OK, then set the .current.textConfig to the requested value indicating that the request executed sucessfull
      {
        dpSet(_sodeObjectDP + ".current.textConfig", textConfig.ToJson());
      }
    }
  }

  /**
    Processes the reply from the getState command
    @param reply
  */
  private void ProcessGetStateReply(shared_ptr<BaseReply> reply)
  {
    bool currentState = false;

    if(reply.GetStatus() == ReplyStatus::OK)
    {
      shared_ptr<GetStateRply> stateReply = (shared_ptr<GetStateRply>)reply;

      if(stateReply.GetState() == "ACTIVE")  //mma: if the getState reply indicates that the sode is ACTIVE then set the correspoinding state dpe to true
      {
        currentState = true;
      }
    }

    dpSet(_sodeObjectDP + ".current.state", currentState);
  }

  /**
    Processes the reply from the getTextConfig command
    @param reply
  */
  private void ProcessGetTextConfigReply(shared_ptr<BaseReply> reply)
  {
    if(reply.GetStatus() == ReplyStatus::OK)
    {
      shared_ptr<GetTextConfigRply> textConfigReply = (shared_ptr<GetTextConfigRply>)reply;

      SodeTextConfig textConfig;

      textConfig.Position = SodeTextConfig::StringToPosition(textConfigReply.GetTextPosition());
      textConfig.Size = textConfigReply.GetTextSize();
      textConfig.Color = textConfigReply.GetTextColor();

      dpSet(_sodeObjectDP + ".current.textConfig", textConfig.ToJson());
    }
  }

//--------------------------------------------------------------------------------
//@private data
//--------------------------------------------------------------------------------
  private shared_ptr<SodeMessageHandler> _connection;
  private string _sodeObjectDP;
};
